package de.hftstuttgart.unifiedticketing.systems.gitlab;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import de.hftstuttgart.unifiedticketing.exceptions.DeserializationException;
import de.hftstuttgart.unifiedticketing.exceptions.HttpResponseException;
import okhttp3.*;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;

import java.io.IOException;
import java.util.*;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

public class GitlabFilterTest
{
    public GitlabFilter instance;
    public Call call;
    public ArgumentCaptor<Request> requestCaptor;
    public Response.Builder responseBuilder;

    @BeforeEach
    public void initBeforeEach()
    {
        GitlabTicketSystem parent = mock(GitlabTicketSystem.class);
        parent.baseUrl = "https://gitlab.some.tld";
        instance = spy(new GitlabFilter(parent));

        OkHttpClient client = mock(OkHttpClient.class);
        call = mock(Call.class);
        requestCaptor = ArgumentCaptor.forClass(Request.class);
        responseBuilder = new Response.Builder()
            .request(new Request.Builder().url("http://test.some.tld/").build())
            .protocol(Protocol.HTTP_1_1)
            .message("some message");

        doReturn(client).when(instance).getHttpClient();
        doReturn(call).when(client).newCall(requestCaptor.capture());
    }

    @Test
    public void testGetServerError() throws IOException
    {
        doReturn(responseBuilder.code(500).build()).when(call).execute();
        assertThrows(HttpResponseException.class, () -> instance.get());
    }

    @Test
    public void testGetClientError() throws IOException
    {
        doReturn(responseBuilder.code(400).build()).when(call).execute();
        assertThrows(HttpResponseException.class, () -> instance.get());
    }

    @Test
    public void testGetNullBody() throws IOException
    {
        doReturn(responseBuilder.code(200).build()).when(call).execute();
        assertThrows(HttpResponseException.class, () -> instance.get());
    }

    @Test
    public void testGetNoJsonBody() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("somestrangething", MediaType.get("application/json")))
                .build())
            .when(call).execute();
        assertThrows(DeserializationException.class, () -> instance.get());
    }

    @Test
    public void testGetRequestParamsPart1() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("[]".getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();

        String[] values = new String[]
            {
                "title contain",
                "5",
                "8",
                "bug",
                "label1",
                "2",
                "50"
            };

        instance
            .withTitleContain(values[0])
            .withId(values[1])
            .withAssigneeId(values[2])
            .withLabel(values[3])
            .withLabel(values[4])
            .setPage(Integer.parseInt(values[5]))
            .setPageSize(Integer.parseInt(values[6]))
            .isOpen()
            .get();

        HttpUrl url = requestCaptor.getValue().url();
        assertAll(
            () -> assertEquals(values[0], url.queryParameter("search")),
            () -> assertEquals(values[1], url.queryParameter("iids")),
            () -> assertEquals(values[2], url.queryParameter("assignee_id")),
            () -> assertEquals(String.format("%s,%s", values[3], values[4]), url.queryParameter("labels")),
            () -> assertEquals(values[5], url.queryParameter("page")),
            () -> assertEquals(values[6], url.queryParameter("per_page")),
            () -> assertEquals("opened", url.queryParameter("state"))
        );
    }

    @Test
    public void testGetRequestParamsPart2() throws IOException
    {
        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create("[]".getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();

        String[] values = new String[]
            {
                "description contain",
                "SomeName"
            };

        instance
            .withDescriptionContain(values[0])
            .withAssigneeName(values[1])
            .isClosed()
            .get();

        HttpUrl url = requestCaptor.getValue().url();
        assertAll(
            () -> assertEquals(values[0], url.queryParameter("search")),
            () -> assertEquals(values[1], url.queryParameter("assignee_username")),
            () -> assertEquals("closed", url.queryParameter("state"))
        );
    }

    @Test
    public void testGetLocalFilter() throws IOException
    {
        GitlabTicketResponse res = new GitlabTicketResponse();
        res.iid = 5;
        res.title = "some title";
        res.description = "descriptive text";
        res.assignees = new LinkedList<>();
        res.state = "open";
        res.labels = new HashSet<>(Arrays.asList("unifiedticketing", "bug"));

        ObjectMapper mapper = new ObjectMapper();
        ArrayNode arrayNode = mapper.createArrayNode();
        arrayNode.add(mapper.valueToTree(res));

        res.iid = 8;
        res.title = "some special title";
        arrayNode.add(mapper.valueToTree(res));

        res.iid = 94;
        res.description = "description with @username marked";
        arrayNode.add(mapper.valueToTree(res));

        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create(arrayNode.toString().getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();

        List<GitlabTicket> result = instance
            .withTitleMatch("^.*special.*$")
            .withDescriptionMatch("^.*@username.*$")
            .get();

        assertAll(
            () -> assertEquals(1, result.size()),
            () -> Assertions.assertEquals("94", result.get(0).getId())
        );
    }

    @Test
    public void testGetDeserialization() throws IOException
    {
        GitlabTicketResponse res = new GitlabTicketResponse();
        res.iid = 99;
        res.title = "some title";
        res.description = "descriptive text";
        res.assignees = new LinkedList<>();
        res.state = "open";
        res.labels = new HashSet<>(Arrays.asList("unifiedticketing", "feature-request"));

        ObjectMapper mapper = new ObjectMapper();
        ArrayNode arrayNode = mapper.createArrayNode();
        arrayNode.add(mapper.valueToTree(res));

        doReturn(
            responseBuilder
                .code(200)
                .body(ResponseBody.create(arrayNode.toString().getBytes(), MediaType.get("application/json")))
                .build())
            .when(call).execute();

        List<GitlabTicket> expected = new LinkedList<>(
            Collections.singleton(GitlabTicket.fromTicketResponse(instance.parent, res)));
        List<GitlabTicket> actual = instance.get();

        assertEquals(expected, actual);
        assertTrue(expected.get(0).deepEquals(actual.get(0)));
    }
}
